/* _decode_file.c
 *
 * By W.H.Scholten 2001
 * This file is released to the public domain
 */


/* For NEXT <file> : */
char prev_unix_filename [ 32 ];

/* The first block has the load/exec address & locked that are used by the
 * cassette system (true for exec? Hmm. should check), so save them.
 */
bool got_first_block = false;
unsigned long load_addr, exec_addr;
unsigned char flag;

void decode_file(char *filename, int save, int save_blocks) {
	int  prev_block = -1;
	struct bbctape_block block;

	block.data = NULL;

	prev_unix_filename [ 0 ] = 0;

	try {
		initial_synch();

		for (;;) { /* until EOF thrown */
			get_block(&block);

			char fn[50];
			if (block.passed_crc) {
				strcpy(fn, block.name);
				bbcim_replace_bad_chars(fn);
			} else {
				fprintf(stderr, "\nblock failed crc\n");
				if (!write_bad_blocks) {
					printf("\n");
					initial_synch();
					continue;
				}
				strcpy(fn, "bad__");
				strcpy(fn+5, block.name);
				bbcim_replace_bad_chars(fn);
			}

			/* Save complete files *******************************************/
			if (save) {
				if (block.number != prev_block + 1) {
					printf(" Expecting block %02X\n", prev_block + 1);
				}
				else {
					prev_block = block.number;

					FILE *fp = fopen(fn, "ab");
					/* Could seek to allow overwriting/different order for rewind! (fill bad blocks with what?) */
					fwrite(block.data, 1, block.length, fp);
					fclose(fp);
					printf(" Writing");
				}
				if (block.number == 0) {
					got_first_block = true;
					load_addr = block.load_address;
					exec_addr = block.exec_address;
					flag = block.flags;

					/* fix NEXT for the previous file */
					if (strlen(prev_unix_filename) > 0) {
						char inf_file[ 32+4 ];
						strcpy(inf_file, prev_unix_filename);
						strcat(inf_file, ".inf");
						FILE *fp = fopen(inf_file, "a");
						if (fp) {
							fprintf(fp, " NEXT %s", fn/*unix_filename*/);
							fclose(fp);
						}
					}

					/* next time this filename: */
					strcpy(prev_unix_filename, fn/*unix_filename*/);
				}

				if ((block.flags & 0x80) && got_first_block) {
					char inf_fn[70];

					prev_block = -1;

					strcpy(inf_fn, fn);
					strcat(inf_fn, ".inf");

					FILE *fp = fopen(inf_fn, "wb");
					fprintf(fp, "TAPE %-10s %8lX %8lX %s", block.name, load_addr, exec_addr, ((flag & 1) ? " Locked" : ""));
					fclose(fp);
				}
				printf("\n");
			}

			/* Save tape blocks **********************************************/
			if (save_blocks) {
				/* WHS: block writing code v3 */
				FILE *fp;;
				char blockname[100];
				int  ii;

				sprintf(blockname, "%s_%02X", fn, block.number);

				for (ii=0; ; ii++) {
					fp = fopen(blockname, "r");
					if (fp == NULL) break; /* new block */


					/* Check wether the block is the same as that of the
					 * existing block file. */
					int  same_block = false;
					int  block_total_length = block.length + block.header_length;
					if (block.length > 0) block_total_length += 2; /* data CRC */

					char tmp_block_buf_orig[ 8192 ];
					memcpy(tmp_block_buf_orig, block.header, block.header_length);
					memcpy(tmp_block_buf_orig + block.header_length, block.data, block_total_length - block.header_length);

					fseek(fp, 0, SEEK_END);
					if (ftell(fp) == block_total_length) {
						fseek(fp, 0, SEEK_SET);
						unsigned char tmp_block_buf [ 8192 ];
						fread(tmp_block_buf, 1, block_total_length, fp);
						if (!memcmp(tmp_block_buf_orig, tmp_block_buf, block_total_length)) {
							same_block = true;
						}
					}
					if (same_block) {
						printf("\n -> block is identical to existing block %s, so skipped", blockname);
						break;
					}

					printf("\n -> block differs from %s", blockname);
					fclose(fp);
					/* Next filename to try: */
					sprintf(blockname, "%s_%d_%02X", fn, ii, block.number);
				}

				if (fp == NULL) { /* Do not overwrite good block files */
					fp = fopen(blockname, "w");
					if (fp == NULL) {
						perror("write error\n");
						exit(1);
					}
					//fwrite(block.data, 1, block_total_length, fp);
					/* write header first */
					fwrite(block.header, 1, block.header_length, fp);
					if (block.length > 0) {
						fwrite(block.data, 1, block.length+2, fp); /* data + CRC */
					}
					printf(" Writing");
				}

				if (block.flags & 0x80) {
					prev_block = -1;
				}
				printf("\n");

				fclose(fp);
			}

			if (!save && !save_blocks) {
				printf("\n");
			}
		}
	}
	catch (GiveUpInputEx &e) {
	}

	if (block.data) free(block.data);
}
