/* Tape image format:
 * first the identifier:
 *      - 8 bytes: "BBCTAPE" + terminating 0
 *
 * followed by data blocks which consist of:
 *  byte  0 : block type:
 *  bytes 1 - 4 : block len (MSB)
 *  bytes 5 - <block_len-1> : data
 * .
 *
 * Currently assigned block types (no other types are needed I think):
 * block type = 0 : silence
 *     - 4 bytes : MSB block length (=1+4+4)
 *     - 4 bytes : MSB duration (ms)
 *
 * block type = 1 : bit stream:
 *     - 4 bytes : MSB block length (=1+4+2+4+ (<nbits> +7)/8)
 *     - 2 bytes : MSB tone frequency (Hz) (= baud rate)
 *     - 4 bytes : MSB number of bits in the block
 *     - < (<nbits> +7)/8 > bytes : data block with the bits
 *
 * block type = 2 : standard BBC tape block
 *     - 4 bytes : MSB block length (=1+4+2+ <tape block size>)
 *     - 2 bytes : MSB tone frequency (Hz) (= baud rate)
 *     - < tape block size > bytes : data block with the actual tape block
 *
 */

FILE *tapeimage_open(char *name) {
	int  st;
	char id[8] = "BBCTAPE";
	char tmp[8];
	FILE *fp = fopen(name, "rb");
	if (fp == NULL) return fp;

	st = fread(tmp, 1, 8, fp);
	if (st != 8 || !memcmp(tmp, id, 8)) {
		fclose(fp);
		return NULL;
	}

	return fp;
}

int tapeimage_next_block(FILE *fp) {

	return -1; /* not yet done */
}


#define BBCTAPE_IMAGE__SILENCE  0
#define BBCTAPE_IMAGE__STREAM   1
#define BBCTAPE_IMAGE__BLOCK    2

void tapeimage_write_header(FILE *fp) {
	char *id = "BBCTAPE";

//	fwrite(id, 1, 8, fp);
	my_fwrite(fp, id, 8, "write_header");
}

void tapeimage_write_stream_header(FILE *fp, unsigned short hz, int length) {
	int i;
	unsigned char mybyte = BBCTAPE_IMAGE__STREAM;
//	int  st;

//	fwrite(&mybyte, 1, 1, fp);
	my_fwrite(fp, &mybyte, 1, "write_stream_header");

	for(i=0; i<4; i++) {
		mybyte = length >> (3-i);
//		st = fwrite(&mybyte, 1, 1, fp);
//		if (st != 1) { perror("write_stream_header"); exit(1); }
		my_fwrite(fp, &mybyte, 1, "write_stream_header");
	}

	for(i=0; i<2; i++) {
		mybyte = hz >> (1-i);
//		fwrite(&mybyte, 1, 1, fp);
		my_fwrite(fp, &mybyte, 1, "write_stream_header");
	}
}

void tapeimage_write_block_header(FILE *fp, unsigned short hz, int length) {
	int i;
	unsigned char mybyte = BBCTAPE_IMAGE__BLOCK;

	fwrite(&mybyte, 1, 1, fp);

	for(i=0; i<4; i++) {
		mybyte = length >> (3-i);
//		fwrite(&mybyte, 1, 1, fp);
		my_fwrite(fp, &mybyte, 1, "write_block_header");
	}

	for(i=0; i<2; i++) {
		mybyte = hz >> (1-i);
//		fwrite(&mybyte, 1, 1, fp);
		my_fwrite(fp, &mybyte, 1, "write_block_header");
	}
}


void initial_synch_whs()
{
	int  i;
	int  dir = 1;

#if 1 /* ORIG maar korter */
	const int flanks = 500;
	CircularBuffer<float> flank(flanks);

	average_flank = 0;

	if (debug) fprintf(stderr, "sync_whs: init\n");

	/* skip first n flanks (but add them to averaging buffer) */
	for (i=0; i<flanks; i++) {
		flank << find_flank(dir); dir = -dir;
		average_flank += flank[0];
	}
	average_flank /= flanks;

	/* keep getting flanks until all flank lengths in buffer are close
	 * "enough" to the average */
	int good_flanks = 0;
	while (good_flanks < flanks) {
		average_flank -= flank[flanks-1] / flanks;
		flank << find_flank(dir);
		dir=-dir;
		average_flank += flank[0] / flanks;
		if (bound(flank[0], float(0.7*average_flank), float(1.3*average_flank)))
			good_flanks++;
		else
			good_flanks = 0;
	}
#else
	const int flanks = 20;
	CircularBuffer<float> flank(flanks);

	average_flank = 0;

	if (debug) fprintf(stderr, "sync_whs v2: init\n");

	for (i=0; i<flanks; i++) {
		flank << find_flank(dir); dir = -dir;
		average_flank += flank[0];
		if (flank[0] > 1.2*average_flank/(i+1) || flank[0] < 0.8*average_flank/(i+1)) {
			//fprintf(stderr, "flank[0]=%f, bounds: %f, %f\n", flank[0], 1.3*average_flank/(i+1), flank[0] < 0.7*average_flank/(i+1));
			i = -1; /* reset */
			average_flank = 0;
		}
	}
	average_flank /= flanks;
#endif


	if (debug) fprintf(stderr, "sync_whs: average flank distance %4.2f achieved at %d\n", average_flank, bytes_read);

	bit_length = 4*average_flank;

	if (debug) fprintf(stderr, "sync_whs: average bit length %4.2f\n", bit_length);

	bit_error = 0;

	low_1 = 0.3*average_flank;
	high_1 = 1.3*average_flank;
	low_0 = 1.3*average_flank;
	high_0 = 3.0*average_flank;
	if (debug) fprintf(stderr, "'1': %4.2f - %4.2f\n", low_1, high_1);
	if (debug) fprintf(stderr, "'0': %4.2f - %4.2f\n", low_0, high_0);
	if (debug) fprintf(stderr, "sync_whs: initial sync done\n");
}


int get_next_bit_whs(FILE *fp)
{
	int start = bytes_read, bit = -1;
int musec;
	int prev_bytes_read = bytes_read;

	find_wave();

	if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
		bit = 0;
	else {
		prev_bytes_read = bytes_read;
		find_wave();
		if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
			bit = 1;
	}

	if (bit < 0) {
		if (debug) fprintf(stderr, "bad bit between %d and %d (%d, %d -> %4.2f, %4.2f)\n", start, bytes_read, prev_bytes_read - start, bytes_read - prev_bytes_read, (prev_bytes_read - start) / bit_length, (bytes_read - prev_bytes_read) / bit_length);
	}
	musec = (bytes_read - start);// / bit_length;
	if (musec > 300) {
		fprintf(stderr, "no signal for %d (%f s)\n", musec, musec / 1.0e6);
		initial_synch_whs(); /* no signal for .1 s */
		int i; unsigned char mybyte = 0xff; for (i=0; i<50; i++) fwrite(&mybyte, 1, 1, fp);

	}
 
	return bit;
}

int get_next_bit_whs_v2(FILE *fp)
{
	int bit = -1;
	int start = bytes_read;
	int prev_bytes_read = bytes_read;

//  do_it_again:
	find_wave();

	if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
		bit = 0;
	else {
		prev_bytes_read = bytes_read;
		find_wave();
		if (bound((bytes_read - start) / bit_length, float(0.80), float(1.2)))
			bit = 1;
		else {
			if (bound((bytes_read - prev_bytes_read) / bit_length, float(0.80), float(1.2))) {
				if (debug) fprintf(stderr, "bad wave (half 1 wave?) at %d: %d -> %4.2f long\n", prev_bytes_read, prev_bytes_read - start, (bytes_read - prev_bytes_read) / bit_length);
//				start = prev_bytes_read;
//				goto do_it_again;
				bit = 0;
			}
		}
	}

	if (bit < 0) {
		if (debug) fprintf(stderr, "bad bit between %d and %d (%d, %d -> %4.2f, %4.2f)\n", start, bytes_read, prev_bytes_read - start, bytes_read - prev_bytes_read, (prev_bytes_read - start) / bit_length, (bytes_read - prev_bytes_read) / bit_length);
	}
 
	return bit;
}




void decode_to_image(char *fn) {
	FILE *fp;
	int bit_no = 0;
	unsigned char mybyte = 0;
	int  do_sync = 0;
//	int  size_fptr;
	int  i;

	fp = fopen(fn, "wb");
	if (fp == NULL) {
		fprintf(stderr, "open %s for writing failed\n", fn);
		exit(1);
	}

	printf("Writing tape image:\n");
//	initial_synch_whs();
	initial_synch();
	mybyte = 0xff; for (i=0; i<500; i++) fwrite(&mybyte, 1, 1, fp);


//	tapeimage_write_header(fp);
//	size_fptr = ftell(fp) + 1; /* to write the stream length later */
//	tapeimage_write_stream_header(fp, 1200, 0); /* length not yet known */


	static int prev_ones = 0;
	static int data_bits = 0;
	for (;;) {
		static int prev_bad_bits;
		int  bit;
		static int  j = 0;

		bit = get_next_bit_whs_v2(fp);

		j++; if (j == 256*8) { j=0; printf("."); fflush(stdout); }

		data_bits++;
		if (bit == 1) {
			prev_ones++;
			if (prev_ones >= 160) {
				/* In a header, still the same one or a new one? */
				if (data_bits >= 800 /*prev_ones*/) {
					fprintf(stderr, "\nmultiple ones (%d) -> header tone detected\n", prev_ones);
					data_bits = 0; prev_ones = 0;
					//do_sync = 1;
				}
				else { /* still in the header tone */
					data_bits = 0; prev_ones = 0;
				}
			}
		} else prev_ones = 0;

		if (bit < 0) {
			prev_bad_bits++;
			//fprintf(stderr, "Bad bit %d\n", prev_bad_bits);
			//if (prev_bad_bits == 4) { prev_bad_bits = 0; do_sync = 1; }
			bit = 1;
		} //else prev_bad_bits = 0;


		mybyte |= bit << bit_no;

		bit_no++;

		if (do_sync) {
			mybyte |= 0xff << bit_no; /* header tone: ones */
			fwrite(&mybyte, 1, 1, fp);
			mybyte = 0;
			bit_no = 0;

			fprintf(stderr, "sync\n");
			initial_synch_whs();
			mybyte = 0xff; for (i=0; i<50; i++) fwrite(&mybyte, 1, 1, fp);

			do_sync = 0;
		} else {
			if (bit_no == 8) {
				//write_bits:
				fwrite(&mybyte, 1, 1, fp);
				mybyte = 0;
				bit_no = 0;
			}
		}
	}
	fclose(fp);
	printf("Done (writing tape image)\n");
}
