/*
 * Copyright (C) W.H.Scholten 2001.
 *
 * Copyright (C) Ole Stauning 1996.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the copyright holder
 * not be used in advertising or publicity pertaining to distribution
 * of the software without specific, written prior permission. The
 * copyright holder makes no representations about the suitability of
 * this software for any purpose. It is provided "as is" without express
 * or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <math.h>

/* Writing sound data: */
//#define SIGNAL_AMP 100
#define SIGNAL_AMP 125
#define SIGNAL_ZERO 128
#define SIGNAL_PHASE -1


int  baud_rate = BBCTAPE_DEFAULT_BAUD_RATE;
int  block_size = BBCTAPE_DEFAULT_BLOCK_SIZE;


/* The crc variable */
u_short _bbctape_crc;

void clear_crc() {
	_bbctape_crc = 0;
}

u_short read_crc() {
	return _bbctape_crc;
}

void byte2crc(u_char b) {
	int i;

	_bbctape_crc ^= b<<8;
	for(i=0;i<8;i++)
		_bbctape_crc = (_bbctape_crc & 0x8000) ? ( (_bbctape_crc ^ 0x810) << 1 ) | 1 : _bbctape_crc << 1;
}


/* the save wave table */
u_char *one, *zero;
int length;

void make_wave_table() {
	int i;

	length = sample_freq / baud_rate;
	if ((zero = (u_char *) malloc(length)) == NULL || (one = (u_char *) malloc(length)) == NULL) {
		fprintf(stderr, "Unable to allocate audio buffer for output\n");
		exit(1);
	}
	for(i=0;i<length;i++) /* zero wave table */
		zero[i] = (u_int8_t) (SIGNAL_ZERO+SIGNAL_AMP*SIGNAL_PHASE*sin((2.0*M_PI*i)/length));
	for(i=0;i<length;i++) /* one wave table  */
		one[i] = zero[(2*i)%length];
}


void delete_wave_table() {
	free(zero);
	free(one);
}


void make_sync(int audio, int sync_length) { /* create the tone */
	int i;

	for(i=0;i<sync_length;i++) {
		if (my_write(audio, one, length) != length) {
			perror(write_file);
			exit(1);
		}
	}
}


void save_byte(int audio, u_char b) {
	int i;
	u_char *onezero;

	byte2crc(b); /* update the crc value */

	for(i=0;i<number_of_start_bits;i++) { /* first write the start bits */
		onezero = (start_bits & (1<<i)) ? one : zero;
		if (my_write(audio, onezero, length) != length) {
			perror(write_file);
			exit(1);
		}
	}
	for(i=0;i<DATA_BITS;i++) { /* then write the data */
		onezero = (b&(1<<i)) ? one : zero;
		if (my_write(audio, onezero, length) != length) {
			perror(write_file);
			exit(1);
		}
	}
	for(i=0;i<number_of_stop_bits;i++) { /* finally write the stop bits */
		onezero = (stop_bits & (1<<i)) ? one : zero;
		if (my_write(audio, onezero, length) != length) {
			perror(write_file);
			exit(1);
		}
	}
}


void save_lsb_long(int audio, u_long ul) {
	int i;
	for(i=0; i<4; i++) {
		save_byte(audio, ul & 0xff);
		ul >>= 8;
	}
}


void save_lsb_short(int audio, u_short us) {
	int i;
	for(i=0; i<2; i++) {
		save_byte(audio, us & 0xff);
		us >>= 8;
	}
}
