/* _uef_extract.c, by W.H.Scholten 2001
 *
 * This file is released to the public domain.
 *
 */

void uef_extract(char *uef_file) {
	unsigned char data[ 8192 ];
	int  next_id;
	int  nfiles = 0;

	int  expect_new_file = true;
	int  block_counter = 0;

	/* Filename on the host system: */
	char unix_filename[ 32 ];

	/* For NEXT <file> : */
	char prev_unix_filename [ 32 ];

	char filename[ 32 ];
	unsigned int  load_addr;
	unsigned int  exec_addr;
	unsigned char flag;
	unsigned int  file_len;

	char block_filename[ 32 ];
	unsigned int  block_load_addr;
	unsigned int  block_exec_addr;
	unsigned int  block_len;
	unsigned int  block_no;
	unsigned char block_flag;
	unsigned int  block_header_crc;
	unsigned int  block_data_crc;

	/* For screen output: */
	char printable_filename[ 32 ];
	char printable_block_filename[ 32 ];

	int chunk_len;

	int tmp_crc;

	bool ok;

	filename[ 0 ] = 0;
	prev_unix_filename [ 0 ] = 0;


	ok = input.Open(uef_file, "r", 8, 0);
	if (ok == 1) {
		fprintf(stderr, "Could not open %s (or not an uef file)\n", uef_file);
		exit(1);
	}

	if (tape_extract)
		printf("Extracting tape blocks to files:\n");
	else if (tape_extract_blocks)
		printf("Extracting tape blocks:\n");
	else if (tape_extract_image)
		printf("Extracting tape data to image:\n");
	else
		printf("Showing tape files:\n");


	DEB(printf("xxx open ok? %d\n", ok));
	do {
		if (input.OverRan()) break;

		next_id = input.GetId();
		DEB(printf("xxx GetId=%d\n", next_id));
		//printf("xxx FindId=%d\n", input.FindId(next_id));
		DEB(printf("xxx GetLength=%d\n", input.GetLength()));
//		input.ReadAll(data);
		switch(next_id) {
		case 0x0000:
			DEB(printf("origin block: %s\n", data));
			break;

		case 0x0100:
		{
			DEB(printf("standard tape block\n"));

			input.ReadAll(data);
			chunk_len = input.GetLength();

			DEB(printf("chunk_len=%d\n", chunk_len));

			if (chunk_len <= 1) {
				DEB(printf("dummy block\n"));
				break;
			}

			strncpy(block_filename, (char *) (data+1), 32);
			block_filename[31] = 0;

			strcpy(printable_block_filename, block_filename);

			bbcim_printable_bbcfilename(printable_block_filename);
//printf("printable_block_filename=%s\n", printable_block_filename);

			unsigned char *d2 = data + 1 + strlen(block_filename) + 1;
			block_load_addr = (*d2) | (*(d2+1) << 8) | (*(d2+2) << 16) | (*(d2+3) << 24);
			d2 += 4;
			block_exec_addr = (*d2) | (*(d2+1) << 8) | (*(d2+2) << 16) | (*(d2+3) << 24);
			d2 += 4;
			block_no = (*d2) | (*(d2+1) << 8);
			d2 += 2;
			block_len = (*d2) | (*(d2+1) << 8);
			d2 += 2;
			block_flag = (*d2);
			DEB(printf("block no=%d, block len=%d\n", block_no, block_len));
			d2 += 1+4; /* skip 4 spare bytes */

			/* 2 byte CRC (MSB): */
			block_header_crc = (*d2 << 8) | (*(d2+1));
			DEB(printf("block header CRC=%4X\n", block_header_crc));
			d2 += 2;


			/* Check the header CRC */
			tmp_crc = bbcim_data_crc(data+1, strlen(block_filename)+1+19 -2);
			if (block_header_crc != tmp_crc) {
				printf("%s: block %d: header CRC check failed (calculated %4X, file %4X)\n", printable_block_filename, block_counter, tmp_crc, block_header_crc);
			}
			else {
				DEB(printf("header CRC ok\n"));
			}

			/* ************ tape image ***************** */
			if (tape_extract_image) {
				printf("%s %02X\n", printable_block_filename, block_no);
				FILE *fp = fopen(tape_image_name, "ab");
				tapeimage_write_block_header(fp, 1200, chunk_len+7);
				my_fwrite(fp, data, chunk_len, tape_image_name);
				fclose(fp);
				break;
			}


			if (!expect_new_file) {
				/* Check that the filename matches */
				if (strlen(filename) > 0 && strcmp(filename, block_filename)) {
					printf("file? (expecting %s %02lX, found %s %02lX)\n", printable_filename, block_counter+1, printable_block_filename, block_no);
					expect_new_file = true;
					block_counter = 0;
				}
			}

			if (expect_new_file) {
				nfiles++; /* Actually found a new file */
				file_len = 0;

				if (block_no != 0) printf("%s: block? (expecting 00, found %02lX)\n", printable_filename, block_no);

				strcpy(filename, block_filename);
				load_addr = block_load_addr;
				exec_addr = block_exec_addr;
				flag = block_flag;

				/* Printable filename */
				strcpy(printable_filename, filename);
				bbcim_printable_bbcfilename(printable_filename);

				/* Local filename */
				strcpy(unix_filename, filename);
				bbcim_replace_bad_chars(unix_filename);
				bbcim_rename_sequential(unix_filename);

				DEB(printf("filename = %s\n", filename));

				if (tape_extract) {
					if (strlen(prev_unix_filename) > 0) {
						char inf_file[ 32+4 ];
						FILE *fp_inf;

						strcpy(inf_file, prev_unix_filename);
						strcat(inf_file, ".inf");

						fp_inf = fopen(inf_file, "a");
						if (fp_inf) {
							fprintf(fp_inf, " NEXT %s", unix_filename);
							fclose(fp_inf);
						}
					}
				}
			}


			/* data: */
			if (tape_extract) {
				FILE *fp = fopen(unix_filename, "a");
				if (fp == NULL) {
					fprintf(stderr, "error opening file\n");
					exit(1);
				}
				fwrite(d2, 1, block_len, fp);
				fclose(fp);
			}

			if (tape_extract_blocks) {
				/* WHS: block writing code v3 */
				FILE *fp;
				char blockname[100];

				sprintf(blockname, "%s_%02lX", unix_filename, block_no);

				int  ii;
				for (ii=0; ; ii++) {

					fp = fopen(blockname, "r");
					if (fp == NULL) break; /* new block */

					/* Check wether the block is the same as that of the
					 * existing block file. */
					int  same_block = false;
					int  block_total_length = chunk_len;

					char tmp_block_buf_orig[ 8192 ];
					memcpy(tmp_block_buf_orig, data, chunk_len);

					fseek(fp, 0, SEEK_END);
					if (ftell(fp) == block_total_length) {
						fseek(fp, 0, SEEK_SET);
						unsigned char tmp_block_buf [ 8192 ];
						fread(tmp_block_buf, 1, block_total_length, fp);
						if (!memcmp(tmp_block_buf_orig, tmp_block_buf, block_total_length)) {
							same_block = true;
						}
					}
					if (same_block) {
						printf("\n -> Identical to existing block %s, so skipped\n", blockname);
						break;
					}

					printf("\n -> block differs from %s\n", blockname);
					fclose(fp);
					/* Next filename to try: */
					sprintf(blockname, "%s_%d_%02X", unix_filename, ii, block_no);
				}

				if (fp == NULL) { /* Do not overwrite good block files */
					fp = fopen(blockname, "w");
					if (fp == NULL) {
						perror("write error\n");
						exit(1);
					}
					fwrite(data, 1, chunk_len, fp);
					//printf(" Writing");
				}
				//printf("\n");

				fclose(fp);
			}

			d2 += block_len;

			file_len += block_len;

			/* 2 byte CRC (MSB), only if #bytes>0: */
			DEB(printf("bytes left: %d\n", chunk_len-1-strlen(filename)-1-4-4-2-2-1-4-2-block_len));
			if (block_len > 0) {
				block_data_crc = (*d2 << 8) | (*(d2+1));
				DEB(printf("block data CRC=%4X\n", block_data_crc));
				d2 += 2;

				/* Check the data CRC */
				tmp_crc = bbcim_data_crc(data+1+strlen(filename)+1+19 -2 +2, block_len);
				if (block_data_crc != tmp_crc) {
					printf("%s: block %d: data CRC check failed (calculated %4X, file %4X)\n", printable_filename, block_counter, tmp_crc, block_data_crc);
				}
				else {
					DEB(printf("data CRC ok\n"));
				}
			}


			if (expect_new_file) {
				if (tape_extract) {
					FILE *fp_inf;
					char inf_file[ 32+4 ];

					strcpy(inf_file, unix_filename);
					strcat(inf_file, ".inf");

					fp_inf = fopen(inf_file, "w");
					if (fp_inf == NULL) {
						fprintf(stderr, "error opening .inf file\n");
						exit(1);
					}
					fprintf(fp_inf, "TAPE %-10s %8X %8X %sCRC=%4X", filename, load_addr, exec_addr, (flag & 1 ? "Locked " : ""), bbcim_fcrc(unix_filename, 1));
					fclose(fp_inf);
				}

				expect_new_file = false;
			} else {
				/* check for inconsistencies */
				if (strcmp(filename, block_filename)) {
					printf("file? (%s -> %s)\n", printable_filename, block_filename);
				}
				/* Should also check change of load address/exec address */
			}

			if (block_flag & 0x80) {
				DEB(printf("done\n"));
				if (tape_catalogue) {
					printf("  %-10s %8X %8X %8X %s\n", printable_filename, load_addr, exec_addr, file_len, flag & 1 ? "Locked " : "");
				}
				expect_new_file = true;
				block_counter = 0;

				strcpy(prev_unix_filename, unix_filename);
			} else {
				block_counter++;
			}
			break;
		}

		case 0x0110:
		case 0x0111:
		case 0x0112:
			DEB(printf("tape data (high tone/gap etc.)\n"));
			break;

		default:
			fprintf(stderr, "unhandled block %d\n", next_id);
			break;
		}
		DEB(printf("\n"));
		if (input.OverRan()) break;
	}	while (!input.FindNext());

	if (nfiles == 0) printf("(No files in the image)\n");
}
